import 'dart:async';

import 'package:build/build.dart';
import 'package:path/path.dart';
import 'package:source_gen/source_gen.dart';
import 'package:grpc/service_api.dart' as grpc;

class ServiceGenerator extends Generator {
  @override
  FutureOr<String?> generate(LibraryReader library, BuildStep buildStep) async {
    final imports = [];
    final importBuffer = StringBuffer();
    final bodyBuffer = StringBuffer();

    for (final clazz in library.classes) {
      if (!TypeChecker.fromRuntime(grpc.Service).isSuperOf(clazz)) continue;
      final pathSegments = buildStep.inputId.pathSegments;
      pathSegments.remove("lib");
      final path = joinAll(pathSegments);
      if (!imports.contains(path)) {
        imports.add("package:${buildStep.inputId.package}/$path");
      }

      final baseServiceName = clazz.name;
      final serviceName = baseServiceName.replaceFirst("ServiceBase", "");
      final betterServiceBaseName = "Better${serviceName}ServiceBase";
      final betterServiceName = "Better${serviceName}Service";
      final betterClientBaseName = "Better${serviceName}ClientBase";
      final clientName = "${serviceName}Client";
      final betterClientName = "Better$clientName";
      final offlineClientName = "Offline$clientName";

      /// Generate BetterServiceBase abstract class.
      bodyBuffer.writeln("abstract class $betterServiceBaseName {");
      // bodyBuffer.writeln("@override");
      // bodyBuffer.writeln("String get \$name => '$serviceName';");
      for (final method in clazz.methods) {
        if (!method.isAbstract) continue;
        bodyBuffer.writeln(
            "${method.returnType.toString().replaceFirst('Future', 'FutureOr')} ${method.name} (${method.parameters.last});");

        // bodyBuffer.write(
        //     "${method.returnType} ${method.name} (${method.parameters.join(',')})");
        // if (method.returnType.isDartAsyncFuture) bodyBuffer.write(" async ");
        // bodyBuffer.writeln("=>");
        // if (method.returnType.isDartAsyncFuture) bodyBuffer.write(" await ");
        // bodyBuffer.writeln("${method.name}Request(request);");
      }
      bodyBuffer.writeln(
          "$betterServiceName get service => $betterServiceName(this);");
      bodyBuffer.writeln("}");

      /// Generate BetterService abstract class.
      bodyBuffer.writeln("class $betterServiceName extends $baseServiceName {");
      bodyBuffer.writeln("final $betterServiceBaseName _service;");
      bodyBuffer.writeln("$betterServiceName(this._service);");
      // bodyBuffer.writeln("@override");
      // bodyBuffer.writeln("String get \$name => '$serviceName';");
      for (final method in clazz.methods) {
        if (!method.isAbstract) continue;
        // bodyBuffer.writeln(
        //     "${method.returnType} ${method.name} (${method.parameters.join(',')}) => _service.${method.name}(request);");
        bodyBuffer.writeln("@override");
        bodyBuffer.write(
            "${method.returnType} ${method.name} (${method.parameters.join(',')})");
        if (method.returnType.isDartAsyncFuture) bodyBuffer.write(" async ");
        bodyBuffer.writeln("=>");
        if (method.returnType.isDartAsyncFuture) bodyBuffer.write(" await ");
        bodyBuffer.writeln("_service.${method.name}(request);");
      }
      bodyBuffer.writeln("}");

      /// Create abstract BetterClientBase class.
      bodyBuffer.writeln("abstract class $betterClientBaseName {");
      for (final method in clazz.methods) {
        if (!method.isAbstract) continue;
        bodyBuffer.writeln(
            "${method.returnType.toString().replaceFirst('Future', 'FutureOr')} ${method.name} (${method.parameters.last});");
      }
      bodyBuffer.writeln("}");

      /// Create normal client BetterClient class.
      bodyBuffer
          .writeln("class $betterClientName extends $betterClientBaseName {");
      bodyBuffer.writeln("final $clientName _client;");
      bodyBuffer.writeln("$betterClientName(this._client);");
      for (final method in clazz.methods) {
        if (!method.isAbstract) continue;
        bodyBuffer.writeln("@override");
        bodyBuffer.writeln(
            "${method.returnType.toString().replaceFirst('Future', 'FutureOr')} ${method.name} (${method.parameters.last}) => _client.${method.name}(request);");
      }
      bodyBuffer.writeln("}");

      /// Create offline client BetterClient class.
      bodyBuffer
          .writeln("class $offlineClientName extends $betterClientBaseName {");
      bodyBuffer
          .writeln("final $betterServiceBaseName _betterSystemServiceBase;");
      bodyBuffer.writeln("$offlineClientName(this._betterSystemServiceBase);");
      for (final method in clazz.methods) {
        if (!method.isAbstract) continue;
        bodyBuffer.writeln("@override");
        bodyBuffer.writeln(
            "${method.returnType.toString().replaceFirst('Future', 'FutureOr')} ${method.name} (${method.parameters.last}) => _betterSystemServiceBase.${method.name}(request);");
      }
      bodyBuffer.writeln("}");
    }

    /// add imports.
    imports.add("dart:async");
    imports.add("package:grpc/service_api.dart");

    /// toSet().toList() to remove duplicated imports
    for (final import in imports.toSet().toList()) {
      importBuffer.writeln("import '$import';");
    }

    /// Return the results.
    final buffer = StringBuffer();
    buffer.write(importBuffer);
    buffer.write(bodyBuffer);
    if (bodyBuffer.isEmpty) return null;
    return buffer.toString();
  }
}
